#include "FixedBuffer.h"

#include "../pointers/pointers.h"

#include <stdlib.h>
#include <string.h>

int FixedBuffer_Create(FixedBuffer* target, size_t size, allocator_t* allocator)
{
	if (target == NULL) return EDESTADDRREQ;

	void* memory = Allocator_Allocate(allocator, size);
	if (memory == NULL) return ENOMEM;

	target->memory = memory;
	target->length = size;
	target->reserved = 0;

	return EXIT_SUCCESS;
}

void* FixedBuffer_GetHead(FixedBuffer* buffer)
{
	if (buffer == NULL) return NULL;

	return advancep(buffer->memory, buffer->reserved);
}

size_t FixedBuffer_GetRemaining(FixedBuffer* buffer)
{
	if (buffer == NULL) return EINVAL;

	return buffer->length - buffer->reserved;
}

int FixedBuffer_DiscardTail(FixedBuffer* buffer, size_t length)
{
	if (buffer == NULL) return EINVAL;
	if (length > buffer->reserved) return EBOUNDS;

	memmove(buffer->memory, advancep(buffer->memory, length), buffer->reserved - length);
	buffer->reserved -= length;

	return EXIT_SUCCESS;
}

int FixedBuffer_Rewind(FixedBuffer* buffer, size_t by)
{
	if (buffer == NULL) return EINVAL;
	if (buffer->reserved < by) return EBOUNDS;

	buffer->reserved -= by;

	return EXIT_SUCCESS;
}

int FixedBuffer_Advance(FixedBuffer* buffer, size_t by)
{
	if (buffer == NULL) return EINVAL;
	if (buffer->reserved + by  > buffer->length) return EBOUNDS;

	buffer->reserved += by;

	return EXIT_SUCCESS;
}

int FixedBuffer_Reset(FixedBuffer* buffer)
{
	if (buffer == NULL) return EINVAL;

	buffer->reserved = 0;

	return EXIT_SUCCESS;
}

void FixedBuffer_Destroy(FixedBuffer* buffer, allocator_t* allocator)
{
	Allocator_Free(allocator, buffer->memory, buffer->length);

	return;
}
