#ifndef OS_SEMAPHORE_H
#define OS_SEMAPHORE_H

#if defined(__unix__) || (defined(__APPLE__) && defined(__MACH__)) // Fuck you, apple

#include <semaphore.h>
typedef sem_t os_semaphore_t;

#elif defined(_WIN32)

#include <windows.h>
typedef struct Struct_OSSemaphore {
	atomic int	count;
	HANDLE		sem_handle;
} os_semaphore_t;

#else
#error "Unknown Threading Standard, feel free to add support via a fork/PR"
#endif

/*!
   @brief Create a Operating-System-Backed Semaphore
   @param target Memory area for the semaphore
	 @param initialValue Value the semaphore will initially hold
   @pre target must be a valid pointer
   @return EXIT_SUCCESS, ENOMEM
*/
int OSSemaphore_Create(os_semaphore_t* target, unsigned int initialValue);

/*!
   @brief Rertieve the value of the semaphore
   @param semaphore this
   @pre Must be a valid semaphore
   @return the value of the semaphore, negative if there are threads waiting
*/
int OSSemaphore_GetValue(os_semaphore_t* semaphore);

/*!
   @brief Wait for a resource to be available on the semaphore
   @param semaphore this
   @return EXIT_SUCCESS, EINVAL, EINTR
*/
int OSSemaphore_Wait(os_semaphore_t* semaphore);

/*!
   @brief Signal the release of a resource on the semaphore
   @param semaphore this
   @return EINVAL, EOVERFLOW
*/
int OSSemaphore_Release(os_semaphore_t* semaphore);

/*!
   @brief Destroy the referenced semaphore resource and return it to the OS
   @param semaphore this
   @return void
*/
void OSSemaphore_Destroy(os_semaphore_t* semaphore);

#endif
